/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */

/*
  This is a simple zero order hold / drop sample resampler.  It's low
  quality, but it's fast and it was easy to write.  It would be nice
  if it filtered out frequencies above the Nyquist frequency of the
  target sample rate, in the case of resampling to a lower rate, or
  the original sample rate, in the case of resampling to a higher
  rate.
*/

#include <stdio.h>
#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>

#include "config.h"
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "main.h"
#include "resampler.h"

typedef struct {
  float conversion;  /* this is actual samples per virual sample */
  int16_t cur_val;
  float remainder;
  int needs_first_read;
} resampler_state;

extern int dying;

void init_resampler(resampler_state *this, int actual_rate,
                    int virtual_rate)
{

  assert(actual_rate > 0);
  assert(virtual_rate > 0);

  this->conversion = (float)actual_rate / (float)virtual_rate;
  this->cur_val = 0;
  this->remainder = 0;
  this->needs_first_read = TRUE;

  assert(this->conversion > 0);
}


void resample(int actual_rate,
              int virtual_rate,
              const int16_t *inbuf,
              int16_t **outbuf,
              int len,
              int *new_len)
{
  int i;
  int in_pos = 0;
  resampler_state this;
  int virtual_samples;

  init_resampler(&this, actual_rate, virtual_rate);
  virtual_samples = (int)(((float)len) / this.conversion);

  *new_len = virtual_samples;

  *outbuf = (int16_t *)malloc(virtual_samples * sizeof(int16_t));
  if (*outbuf == NULL) {
    fprintf(stderr, "couldn't allocate %d bytes for resampled source buffer\n",
            virtual_samples * sizeof(int16_t));
    exit(0);
  }

  for (i = 0; i < virtual_samples; i++) {
    if (dying) {
      return;
    }

    this.remainder += this.conversion;

    if (this.remainder >= 1 || this.needs_first_read) {
      int real_samples = (int)this.remainder;

      if (in_pos >= len) {
        return;
      }

      if (real_samples < 1) {
        assert(this.needs_first_read);
        this.cur_val = inbuf[in_pos];
        in_pos++;
      } else {
        this.cur_val = inbuf[in_pos];
        in_pos += real_samples;

        this.remainder -= real_samples;
      }

      this.needs_first_read = FALSE;

      (*outbuf)[i] = this.cur_val;
    } else {
      /* cur_val hasn't just been updated, so we need to interpolate.
         This is a zero-order hold resampler, so we just hold the
         previous sample value. */

      (*outbuf)[i] = this.cur_val;
    }
  }

  return;
}
