/* soundmosaic, Copyright (c) 2001-2007 Steven Hazel <sah@thalassocracy.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 */
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <limits.h>
#include <audiofile.h>
#include <math.h>
#include <assert.h>
#include <stdlib.h>

#include "sm_utils.h"
#include "soundmosaic.h"

soundmosaic_state *soundmosaic_init(int16_t **src, int32_t src_len,
                                    int src_channels,
                                    void (*draw_match)(int16_t *target_tile,
                                                       int16_t *source_tile,
                                                       int32_t tile_size))
{
  soundmosaic_state *self;

  self = malloc(sizeof(soundmosaic_state));
  if (self == NULL) {
    return NULL;
  }

  self->src = src;
  self->src_len = src_len;
  self->src_channels = src_channels;
  self->draw_match = draw_match;

  assert(self->src_len > 0);
  assert(self->src_channels > 0);

  return self;
}


void soundmosaic_free(soundmosaic_state *self)
{
  if (self != NULL) {
    free(self);
  }
}


int soundmosaic(const soundmosaic_state *self,
                int16_t *tgtbuffer,
                int16_t *outbuffer,
                int32_t tile_size,
                int partition)
{
  int16_t *srctile;
  int16_t temp_src_tile[tile_size];
  int16_t temp_tgt_tile[tile_size];
  int64_t highest, corr;
  uint32_t i, j;
  int64_t tgtmag, tgtmean;
  int32_t tile_spacing;

  assert(self->src_len > 0);
  assert(self->src_channels > 0);

  assert(tile_size > 0);
  if (tile_size < 1) {
    return SOUNDMOSAIC_INVALID_TILE_SIZE;
  }

  /* make sure tile_size won't overflow the final memcpy */
  if (((int64_t)tile_size * (int64_t)sizeof(int16_t)) > INT32_MAX) {
    return SOUNDMOSAIC_INVALID_TILE_SIZE;
  }

  if (tile_size > self->src_len) {
    return SOUNDMOSAIC_INVALID_TILE_SIZE;
  }

  if (partition) {
    tile_spacing = tile_size;
  } else {
    tile_spacing = 1;
  }

  /*fprintf(stderr, "length for this tile is: %ld\n", tile_size);*/

  tgtmag = magnitude(tgtbuffer, tile_size);
  tgtmean = mean(tgtbuffer, tile_size);

  highest = 0;

  for (i = 0; i < self->src_channels; i++) {

    for (j = 0;
         j <= (self->src_len - tile_size);
         j += tile_spacing) {

      srctile = &(self->src[i][j]);

      mean_shift(srctile, temp_src_tile, tile_size, 0);
      mean_shift(tgtbuffer, temp_tgt_tile, tile_size, 0);

      scale_mag(temp_src_tile, temp_src_tile, tile_size, tgtmag);

      corr = correlation(temp_tgt_tile, temp_src_tile, tile_size);

      /*corr = llabs(corr);*/

      if (corr > highest) {
        /*
          fprintf(stderr, "%d: %lu < %lu\n", i,
                  (unsigned long int)corr, (unsigned long int)highest);
        */
        highest = corr;

        mean_shift(temp_src_tile, temp_src_tile, tile_size, tgtmean);
        memcpy(outbuffer, temp_src_tile, (tile_size * sizeof(int16_t)));

        if (self->draw_match != NULL) {
          (*self->draw_match)(tgtbuffer, temp_src_tile, tile_size);
        }
      }
    }
  }

  /*
  fprintf(stderr, "highest correlation is %lu\n",
          (unsigned long int)highest);;
  fprintf(stderr, "final tested source tile was %d\n", i);
  */

  return SOUNDMOSAIC_SUCCESS;
}


int soundmosaic_collate(const int16_t *target,
                        const int16_t *src1,
                        const int16_t *src2,
                        int16_t *output,
                        int32_t tile_size)
{
  int64_t highest, corr;

  /* make sure tile_size won't overflow the final memcpy */
  if (((int64_t)tile_size * (int64_t)sizeof(int16_t)) > INT32_MAX) {
    return SOUNDMOSAIC_INVALID_TILE_SIZE;
  }

  /*fprintf(stderr, "length for this tile is: %ld\n", tile_size);*/

  highest = 0;

  corr = correlation(target, src1, tile_size);

  highest = corr;
  memcpy(output, src1, tile_size * sizeof(int16_t));

  corr = correlation(target, src2, tile_size);

  /*corr = llabs(corr);*/

  if (corr > highest) {
    /*fprintf(stderr, "%lu < %lu\n", (long int)dif, (long int)closest);
      fprintf(stderr, "second file is closer\n");*/
    highest = corr;
    memcpy(output, src2, tile_size * sizeof(int16_t));
  } else {
    /*fprintf(stderr, "first file is closer\n");*/
  }

  /*fprintf(stderr, "closest match was %lu points away\n",
            (unsigned long int)closest);*/

  return SOUNDMOSAIC_SUCCESS;
}
